# standard libraries
import logging
import unittest

# third party libraries
# None

# local libraries
# None


class TestApplicationClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()
